@@ -6,34 +6,30 @@
 #include <stdio.h>
 #include <stdlib.h>
 
-#define CHINADNS_VERSION "ChinaDNS-NG 2023.03.10 <https://github.com/zfl9/chinadns-ng>"
-
-/* limits.h */
-#ifndef PATH_MAX
-  #define PATH_MAX 4096 /* include \0 */
-#endif
+#define CHINADNS_VERSION "ChinaDNS-NG 2023.04.11-2 <https://github.com/zfl9/chinadns-ng>"
 
 bool    g_verbose       = false;
 bool    g_reuse_port    = false;
 bool    g_noip_as_chnip = false; /* default: see as not-china-ip */
-uint8_t g_noaaaa_query  = 0; /* disable AAAA query (bit flags) */
-uint8_t g_default_tag   = NAME_TAG_NONE;
+u8      g_noaaaa_query  = 0; /* disable AAAA query (bit flags) */
+u8      g_default_tag   = NAME_TAG_NONE;
 
-const char *g_gfwlist_fname = NULL; /* gfwlist filename */
-const char *g_chnlist_fname = NULL; /* chnlist filename */
+const char *g_gfwlist_fname = NULL; /* gfwlist filename(s) "a.txt,b.txt,..." */
+const char *g_chnlist_fname = NULL; /* chnlist filename(s) "m.txt,n.txt,..." */
 bool        g_gfwlist_first = true; /* match gfwlist first */
+bool        g_add_tagchn_ip = false; /* add the answer ip of name-tag:chn to ipset/nftset */
 
-char g_ipset_setname4[IPSET_MAXNAMELEN] = "chnroute"; /* ipset setname for ipv4 */
-char g_ipset_setname6[IPSET_MAXNAMELEN] = "chnroute6"; /* ipset setname for ipv6 */
+const char *g_ipset_name4 = "chnroute"; /* ipset:"set_name" | nftset:"family_name@table_name@set_name" */
+const char *g_ipset_name6 = "chnroute6"; /* ipset:"set_name" | nftset:"family_name@table_name@set_name" */
 
-char     g_bind_ipstr[INET6_ADDRSTRLEN] = "127.0.0.1";
-portno_t g_bind_portno                  = 65353;
-skaddr_u g_bind_skaddr;
+const char     *g_bind_ip   = "127.0.0.1";
+u16             g_bind_port = 65353;
+union skaddr    g_bind_skaddr;
 
-char     g_remote_ipports[SERVER_MAXCNT][ADDRPORT_STRLEN] = {[CHINADNS1_IDX] = "114.114.114.114", [TRUSTDNS1_IDX] = "8.8.8.8"};
-skaddr_u g_remote_skaddrs[SERVER_MAXCNT];
-int      g_upstream_timeout_sec                           = 5;
-uint8_t  g_repeat_times                                   = 1; /* used by trust-dns only */
+const char     *g_upstream_addrs[SERVER_MAXCNT];
+union skaddr    g_upstream_skaddrs[SERVER_MAXCNT];
+int             g_upstream_timeout_sec = 5;
+u8              g_repeat_times         = 1; /* used by trust-dns only */
 
 #define OPT_BIND_ADDR 'b'
 #define OPT_BIND_PORT 'l'
@@ -47,6 +43,7 @@ uint8_t  g_repeat_times                                   = 1; /* used by trust-
 #define OPT_TIMEOUT_SEC 'o'
 #define OPT_REPEAT_TIMES 'p'
 #define OPT_CHNLIST_FIRST 'M'
+#define OPT_ADD_TAGCHN_IP 'a'
 #define OPT_NO_IPV6 'N'
 #define OPT_FAIR_MODE 'f'
 #define OPT_REUSE_PORT 'r'
@@ -70,6 +67,7 @@ static const char s_shortopts[] = {
     OPT_REPEAT_TIMES, ':', /* required_argument */
     OPT_NO_IPV6, ':', ':', /* optional_argument */
     OPT_CHNLIST_FIRST, /* no_argument */
+    OPT_ADD_TAGCHN_IP, /* no_argument */
     OPT_FAIR_MODE, /* no_argument */
     OPT_REUSE_PORT, /* no_argument */
     OPT_NOIP_AS_CHNIP, /* no_argument */
@@ -93,6 +91,7 @@ static const struct option s_options[] = {
     {"repeat-times",  required_argument, NULL, OPT_REPEAT_TIMES},
     {"no-ipv6",       optional_argument, NULL, OPT_NO_IPV6},
     {"chnlist-first", no_argument,       NULL, OPT_CHNLIST_FIRST},
+    {"add-tagchn-ip", no_argument,       NULL, OPT_ADD_TAGCHN_IP},
     {"fair-mode",     no_argument,       NULL, OPT_FAIR_MODE},
     {"reuse-port",    no_argument,       NULL, OPT_REUSE_PORT},
     {"noip-as-chnip", no_argument,       NULL, OPT_NOIP_AS_CHNIP},
@@ -110,11 +109,13 @@ static void show_help(void) {
            " -t, --trust-dns <ip[#port],...>      trust dns server, default: <GoogleDNS>\n"
            " -4, --ipset-name4 <ipv4-setname>     ipset ipv4 set name, default: chnroute\n"
            " -6, --ipset-name6 <ipv6-setname>     ipset ipv6 set name, default: chnroute6\n"
-           " -g, --gfwlist-file <file-path>       filepath of gfwlist, '-' indicate stdin\n"
-           " -m, --chnlist-file <file-path>       filepath of chnlist, '-' indicate stdin\n"
+           "                                      if it contains @, then use nftables set\n"
+           "                                      format: family_name@table_name@set_name\n"
+           " -g, --gfwlist-file <path,...>        path(s) of gfwlist, '-' indicate stdin\n"
+           " -m, --chnlist-file <path,...>        path(s) of chnlist, '-' indicate stdin\n"
            " -d, --default-tag <name-tag>         domain default tag: gfw,chn,none(default)\n"
            " -o, --timeout-sec <query-timeout>    timeout of the upstream dns, default: 5\n"
-           " -p, --repeat-times <repeat-times>    it is only used for trustdns, default: 1\n"
+           " -p, --repeat-times <repeat-times>    only used for trustdns, default:1, max:5\n"
            " -N, --no-ipv6=[rules]                filter AAAA query, rules can be a seq of:\n"
            "                                      rule a: filter all domain name (default)\n"
            "                                      rule g: filter the name with tag gfw\n"
@@ -123,8 +124,10 @@ static void show_help(void) {
            "                                      rule c: do not forward to china upstream\n"
            "                                      rule t: do not forward to trust upstream\n"
            "                                      rule C: check answer ip of china upstream\n"
-           "                                      if no rules is given, it defaults to a\n"
+           "                                      rule T: check answer ip of trust upstream\n"
+           "                                      if no rules is given, it defaults to 'a'\n"
            " -M, --chnlist-first                  match chnlist first, default: <disabled>\n"
+           " -a, --add-tagchn-ip                  add the ip of name-tag:chn to ipset/nftset\n"
            " -f, --fair-mode                      enable fair mode (nop, only fair mode now)\n"
            " -r, --reuse-port                     enable SO_REUSEPORT, default: <disabled>\n"
            " -n, --noip-as-chnip                  accept reply without ipaddr (A/AAAA query)\n"
@@ -141,35 +144,51 @@ static void show_help(void) {
     exit(1); \
 })
 
-static void parse_upstream_addrs(char *arg, bool is_chinadns) {
+static void parse_upstream_addrs(const char *arg, bool is_chinadns) {
     int cnt = 0;
+    int has_next = 1;
 
-    for (char *ipstr = strtok(arg, ","); ipstr; ipstr = strtok(NULL, ",")) {
+    do {
         if (++cnt > SERVER_GROUP_CNT)
             err_exit("%s dns servers max count is %d", is_chinadns ? "china" : "trust", SERVER_GROUP_CNT);
 
-        portno_t port = 53;
+        const char *d = strchr(arg, ',');
+        size_t len = d ? (size_t)(d - arg) : strlen(arg);
+
+        if (len + 1 > IP_PORT_STRLEN)
+            err_exit("server addr max length is %d: %.*s", IP_PORT_STRLEN - 1, (int)len, arg);
+
+        /* tmp buffer */
+        char ipstr[IP_PORT_STRLEN];
+        memcpy(ipstr, arg, len);
+        ipstr[len] = '\0';
+
+        /* g_upstream_addrs */
+        char *addr = malloc(len + 1);
+        memcpy(addr, arg, len);
+        addr[len] = '\0';
+
+        if (d)
+            arg += len + 1;
+        else
+            has_next = 0;
+
+        u16 port = 53;
         char *port_str = strchr(ipstr, '#');
         if (port_str) {
             *port_str++ = 0;
-            if (strlen(port_str) + 1 > PORTSTR_MAXLEN)
-                err_exit("port number max length is %d: %s", PORTSTR_MAXLEN - 1, port_str);
-            port = strtoul(port_str, NULL, 10);
-            if (port == 0)
+            if ((port = strtoul(port_str, NULL, 10)) == 0)
                 err_exit("invalid server port number: %s", port_str);
         }
 
-        if (strlen(ipstr) + 1 > INET6_ADDRSTRLEN)
-            err_exit("ip address max length is %d: %s", INET6_ADDRSTRLEN - 1, ipstr);
-
         int family = get_ipstr_family(ipstr);
         if (family == -1)
             err_exit("invalid server ip address: %s", ipstr);
 
         int idx = (is_chinadns ? CHINADNS1_IDX : TRUSTDNS1_IDX) + cnt - 1;
-        sprintf(g_remote_ipports[idx], "%s#%u", ipstr, (uint)port);
-        build_socket_addr(family, &g_remote_skaddrs[idx], ipstr, port);
-    }
+        skaddr_build(family, &g_upstream_skaddrs[idx], ipstr, port);
+        g_upstream_addrs[idx] = addr;
+    } while (has_next);
 }
 
 static void parse_noaaaa_rules(const char *rules) {
@@ -207,6 +226,9 @@ static void parse_noaaaa_rules(const char *rules) {
             case 'C':
                 g_noaaaa_query |= NOAAAA_CHINA_IPCHK;
                 break;
+            case 'T':
+                g_noaaaa_query |= NOAAAA_TRUST_IPCHK;
+                break;
             default:
                 err_exit("invalid no-aaaa rule: '%c'", *c);
                 break;
@@ -215,9 +237,11 @@ static void parse_noaaaa_rules(const char *rules) {
 
     /* try simplify to NOAAAA_ALL */
     if (!is_filter_all_v6(g_noaaaa_query)) {
-        if ((g_noaaaa_query & NOAAAA_TAG_GFW) && (g_noaaaa_query & NOAAAA_TAG_CHN) && (g_noaaaa_query & NOAAAA_TAG_NONE))
+        u8 all_tag = NOAAAA_TAG_GFW|NOAAAA_TAG_CHN|NOAAAA_TAG_NONE;
+        u8 all_dns = NOAAAA_CHINA_DNS|NOAAAA_TRUST_DNS;
+        if ((g_noaaaa_query & all_tag) == all_tag)
             g_noaaaa_query = NOAAAA_ALL;
-        else if ((g_noaaaa_query & NOAAAA_CHINA_DNS) && (g_noaaaa_query & NOAAAA_TRUST_DNS))
+        else if ((g_noaaaa_query & all_dns) == all_dns)
             g_noaaaa_query = NOAAAA_ALL;
     }
 }
@@ -228,51 +252,46 @@ void opt_parse(int argc, char *argv[]) {
     int optindex = -1;
     int shortopt = -1;
 
-    const char *chinadns_optarg = NULL;
-    const char *trustdns_optarg = NULL;
+    const char *chinadns_optarg = "114.114.114.114";
+    const char *trustdns_optarg = "8.8.8.8";
 
     while ((shortopt = getopt_long(argc, argv, s_shortopts, s_options, &optindex)) != -1) {
         switch (shortopt) {
             case OPT_BIND_ADDR:
-                if (strlen(optarg) + 1 > INET6_ADDRSTRLEN)
-                    err_exit("ip address max length is %d: %s", INET6_ADDRSTRLEN - 1, optarg);
                 if (get_ipstr_family(optarg) == -1)
                     err_exit("invalid listen ip address: %s", optarg);
-                strcpy(g_bind_ipstr, optarg);
+                g_bind_ip = optarg;
                 break;
+
             case OPT_BIND_PORT:
-                if (strlen(optarg) + 1 > PORTSTR_MAXLEN)
-                    err_exit("port number max length is %d: %s", PORTSTR_MAXLEN - 1, optarg);
-                g_bind_portno = strtoul(optarg, NULL, 10);
-                if (g_bind_portno == 0)
+                if ((g_bind_port = strtoul(optarg, NULL, 10)) == 0)
                     err_exit("invalid listen port number: %s", optarg);
                 break;
+
             case OPT_CHINA_DNS:
                 chinadns_optarg = optarg;
                 break;
+
             case OPT_TRUST_DNS:
                 trustdns_optarg = optarg;
                 break;
+
             case OPT_IPSET_NAME4:
-                if (strlen(optarg) + 1 > IPSET_MAXNAMELEN)
-                    err_exit("ipset setname max length is %d: %s", IPSET_MAXNAMELEN - 1, optarg);
-                strcpy(g_ipset_setname4, optarg);
+                g_ipset_name4 = optarg;
                 break;
+
             case OPT_IPSET_NAME6:
-                if (strlen(optarg) + 1 > IPSET_MAXNAMELEN)
-                    err_exit("ipset setname max length is %d: %s", IPSET_MAXNAMELEN - 1, optarg);
-                strcpy(g_ipset_setname6, optarg);
+                g_ipset_name6 = optarg;
                 break;
+
             case OPT_GFWLIST_FILE:
-                if (strlen(optarg) + 1 > PATH_MAX)
-                    err_exit("file path max length is %d: %s", PATH_MAX - 1, optarg);
                 g_gfwlist_fname = optarg;
                 break;
+
             case OPT_CHNLIST_FILE:
-                if (strlen(optarg) + 1 > PATH_MAX)
-                    err_exit("file path max length is %d: %s", PATH_MAX - 1, optarg);
                 g_chnlist_fname = optarg;
                 break;
+
             case OPT_DEFAULT_TAG:
                 if (strcmp(optarg, "gfw") == 0)
                     g_default_tag = NAME_TAG_GFW;
@@ -283,46 +302,61 @@ void opt_parse(int argc, char *argv[]) {
                 else
                     err_exit("invalid default domain tag: %s", optarg);
                 break;
+
             case OPT_TIMEOUT_SEC:
-                g_upstream_timeout_sec = strtoul(optarg, NULL, 10);
-                if (g_upstream_timeout_sec <= 0)
+                if ((g_upstream_timeout_sec = strtoul(optarg, NULL, 10)) <= 0)
                     err_exit("invalid upstream timeout sec: %s", optarg);
                 break;
+
             case OPT_REPEAT_TIMES:
-                g_repeat_times = strtoul(optarg, NULL, 10);
-                if (g_repeat_times == 0)
+                if ((g_repeat_times = strtoul(optarg, NULL, 10)) == 0)
                     err_exit("invalid trustdns repeat times: %s", optarg);
+                g_repeat_times = min(g_repeat_times, MAX_REPEAT_TIMES);
                 break;
+
             case OPT_NO_IPV6:
                 parse_noaaaa_rules(optarg);
                 break;
+
             case OPT_CHNLIST_FIRST:
                 g_gfwlist_first = false;
                 break;
+
+            case OPT_ADD_TAGCHN_IP:
+                g_add_tagchn_ip = true;
+                break;
+
             case OPT_FAIR_MODE:
                 /* no operation */
                 break;
+
             case OPT_REUSE_PORT:
                 g_reuse_port = true;
                 break;
+
             case OPT_NOIP_AS_CHNIP:
                 g_noip_as_chnip = true;
                 break;
+
             case OPT_VERBOSE:
                 g_verbose = true;
                 break;
+
             case OPT_VERSION:
                 printf(CHINADNS_VERSION "\n");
                 exit(0);
                 break;
+
             case OPT_HELP:
                 show_help();
                 exit(0);
                 break;
+
             case ':':
                 /* missing argument */
                 err_exit("missing optarg: '%s'", argv[optind - 1]);
                 break;
+
             case '?':
                 /* unknown option */
                 if (optopt) {
@@ -339,24 +373,7 @@ void opt_parse(int argc, char *argv[]) {
         }
     }
 
-    if (g_gfwlist_fname && g_chnlist_fname && strcmp(g_gfwlist_fname, "-") == 0 && strcmp(g_chnlist_fname, "-") == 0)
-        err_exit("gfwlist:%s and chnlist:%s are both STDIN", g_gfwlist_fname, g_chnlist_fname);
-
-    build_socket_addr(get_ipstr_family(g_bind_ipstr), &g_bind_skaddr, g_bind_ipstr, g_bind_portno);
-
-    if (chinadns_optarg) {
-        char buf[strlen(chinadns_optarg) + 1];
-        strcpy(buf, chinadns_optarg);
-        parse_upstream_addrs(buf, true);
-    } else {
-        build_socket_addr(AF_INET, &g_remote_skaddrs[CHINADNS1_IDX], "114.114.114.114", 53);
-    }
-
-    if (trustdns_optarg) {
-        char buf[strlen(trustdns_optarg) + 1];
-        strcpy(buf, trustdns_optarg);
-        parse_upstream_addrs(buf, false);
-    } else {
-        build_socket_addr(AF_INET, &g_remote_skaddrs[TRUSTDNS1_IDX], "8.8.8.8", 53);
-    }
+    skaddr_build(get_ipstr_family(g_bind_ip), &g_bind_skaddr, g_bind_ip, g_bind_port);
+    parse_upstream_addrs(chinadns_optarg, true);
+    parse_upstream_addrs(trustdns_optarg, false);
 }
